# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
# RPi-Spark Audio Device
#
# @author: Kunpeng Zhang
# 2018.4.15
#

import subprocess as sub

class RPiAudioDevice:
    PIN_MODE_AUDIO = "alt0"
    PIN_MODE_OUTPUT = "output"

    ##
    # Audio right channel IO
    channelR = None
    ##
    # Audio left channel IO
    channelL = None
    
    def __init__(self, pinRight, pinLeft):
        """!
        \~english
        Initialize the RPiAudioDevice object instance.
        @param pinRight: Audio right channel IO. "None" means channel disabled
        @param pinLeft: Audio left channel IO. "None" means channel disabled
        @note On Raspberry Pi the audio channel IO can be chosen in ( GPIO in BCM MODE ): 12, 13, 18
        
        \~chinese
        初始化 RPiAudioDevice 对象实例。
        @param pinRight: 音频右声道 IO 。 "None" 表示通道禁用
        @param pinLeft: 音频左声道 IO 。 "None" 表示通道禁用
        @note 树梅派(Raspberry Pi)可选择音频通道输出 IO（BCM 模式）：12,13,18            
        """
        self.channelR = pinRight
        self.channelL = pinLeft
 
    def on(self):
        """!
        \~english
        Open Audio output. set pin mode to ALT0
        @return a boolean value. if True means open audio output is OK otherwise failed to open.

        \~chinese
        打开音频输出。 将引脚模式设置为ALT0
        @return 布尔值。 如果 True 表示打开音频输出成功，否则不成功。
        """
        isOK = True
        try:
            if self.channelR!=None:
                sub.call(["gpio", "-g", "mode", "{}".format(self.channelR), self.PIN_MODE_AUDIO ])
        except:
            isOK = False
            print("Open audio right channel failed.")

        try:
            if self.channelL!=None:
                sub.call(["gpio","-g","mode", "{}".format(self.channelL), self.PIN_MODE_AUDIO ])
        except:
            isOK = False
            print("Open audio left channel failed.")

        return isOK
    
    # Close Audio output. set pin mode to OUTPUT
    def off(self):
        """!
        \~english
        Close Audio output. set pin mode to output
        @return a boolean value. if True means close audio output is OK otherwise failed to close.

        \~chinese
        关闭音频输出。 将引脚模式设置为输出
        @return 布尔值。 如果为 True 关闭音频输出成功，否则关闭不成功。
        """
        isOK = True
        try:
            if self.channelR!=None:
                sub.call(["gpio","-g","mode", "{}".format(self.channelR), self.PIN_MODE_OUTPUT ])
        except:
            isOK = False
            print("Close audio right channel failed.")

        try:
            if self.channelL!=None:
                sub.call(["gpio","-g","mode", "{}".format(self.channelL), self.PIN_MODE_OUTPUT ])
        except:
            isOK = False
            print("Close audio left channel failed.")
        return isOK


# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
# MPU6050
#
# @author: Kunpeng Zhang
# @version v1.0.0
# @date 2018.04.01
#

import smbus

##
# default I2C address of MPU6050
DEF_MPU6050_ADDRESS = 0x68

##
# Earth Gravitiy
GRAVITIY_EARTH  = 9.80665

class MPU6050:
    """!
    \~english
    This driver is for MPU-6050 Motion Processing Unit( InvenSense Inc ).

    Features: <br>
        * Gyroscope sensor
        * Accelerometer sensor
        * Temperature sensor
        * Sensor data ready Interrupt
        * Motion detection Interrupt
        * Sleep mode
        * More...

    \~chinese
    此驱动程序适用于MPU-6050运动处理单元 ( InvenSense Inc )。

         特点：<br>
             * 陀螺仪
             * 加速度计
             * 温度传感器
             * 传感器数据就绪中断
             * 移动侦测中断
             * 睡眠模式
             * 更多...
    """

    # Scale Modifiers
    ACCEL_SCALE_MODIFIER_2G = 16384.0
    ACCEL_SCALE_MODIFIER_4G = 8192.0
    ACCEL_SCALE_MODIFIER_8G = 4096.0
    ACCEL_SCALE_MODIFIER_16G = 2048.0

    GYRO_SCALE_MODIFIER_250DEG = 131.0
    GYRO_SCALE_MODIFIER_500DEG = 65.5
    GYRO_SCALE_MODIFIER_1KDEG = 32.8
    GYRO_SCALE_MODIFIER_2KDEG = 16.4

    # Pre-defined ranges
    ACCEL_RANGE_2G = 0x00
    ACCEL_RANGE_4G = 0x08
    ACCEL_RANGE_8G = 0x10
    ACCEL_RANGE_16G = 0x18

    GYRO_RANGE_250DEG = 0x00
    GYRO_RANGE_500DEG = 0x08
    GYRO_RANGE_1KDEG  = 0x10
    GYRO_RANGE_2KDEG  = 0x18

    # MPU-6050 Registers
    REG_PWR_MGMT_1 = 0x6B
    REG_PWR_MGMT_2 = 0x6C
    
    REG_ACCEL_XOUT_H = 0x3B
    REG_ACCEL_YOUT_H = 0x3D
    REG_ACCEL_ZOUT_H = 0x3F

    REG_TEMP_OUT_H = 0x41

    REG_GYRO_XOUT_H = 0x43
    REG_GYRO_YOUT_H = 0x45
    REG_GYRO_ZOUT_H = 0x47

    REG_GYRO_CONFIG     = 0x1B
    REG_ACCEL_CONFIG    = 0x1C

    #Interrupt Register
    REG_INT_PIN_CFG = 0x37      #中断/旁路设置寄存器
    REG_INT_ENABLE  = 0x38      #中断使能寄存器
    REG_INT_STATUS  = 0x3A      #中断状态寄存器

    REG_SIGNAL_PATH_RESET       = 0x68

    #Motion detect registers
    REG_MOTION_DET              = 0x1F     #运动检测阀值设置寄存器
    REG_MOTION_DET_CTRL         = 0x69     #运动检测控制寄存器
    REG_MOTION_DET_DUR          = 0x20

    #INT PIN CFG Values
    #Register 55 – INT Pin / Bypass Enable Configuration INT_PIN_CFG. page 26
    VAL_INT_PIN_CFG_INT_LEVEL       = 0x80
    VAL_INT_PIN_CFG_INT_OPEN        = 0x40
    VAL_INT_PIN_CFG_LATCH_INT_EN    = 0x20
    VAL_INT_PIN_CFG_INT_RD_CLEAR    = 0x10
    VAL_INT_PIN_CFG_FSYNC_INT_LEVEL = 0x08
    VAL_INT_PIN_CFG_FSYNC_INT_EN    = 0x04
    VAL_INT_PIN_CFG_I2C_BYPASS_EN   = 0x02

    #PWR_MGMT_1 Values
    #Register 107(0x6B) – Power Management 1 / PWR_MGMT_1. page 40
    VAL_PWR_MGMT_1_RESET    = 0x80
    VAL_PWR_MGMT_1_SLEEP    = 0x40
    VAL_PWR_MGMT_1_ON_ALL   = 0x00
    VAL_PWR_MGMT_1_ON_CYCLE = 0x20
    VAL_PWR_MGMT_1_OFF_TEMP = 0x08

    VAL_PWR_MGMT_1_CLKSEL_INTERNAL  = 0x00
    VAL_PWR_MGMT_1_CLKSEL_PLL_XG    = 0x01  #PLL with X axis gyroscope reference
    VAL_PWR_MGMT_1_CLKSEL_PLL_YG    = 0x02  #PLL with Y axis gyroscope reference
    VAL_PWR_MGMT_1_CLKSEL_PLL_ZG    = 0x03  #PLL with Z axis gyroscope reference
    VAL_PWR_MGMT_1_CLKSEL_PLL_EXT_32HZ  = 0x04  #PLL with external 32.768kHz reference
    VAL_PWR_MGMT_1_CLKSEL_PLL_EXT_19HZ  = 0x05  #PLL with external 19.2MHz reference
    VAL_PWR_MGMT_1_CLKSEL_STOP_CLK      = 0x07  #Stops the clock and keeps the timing generator in reset

    #PWR_MGMT_2 Values
    VAL_PWR_MGMT_2_LP_WAKE_CTRL_1_25HZ  = 0x00
    VAL_PWR_MGMT_2_LP_WAKE_CTRL_5HZ     = 0x40
    VAL_PWR_MGMT_2_LP_WAKE_CTRL_20HZ    = 0x80
    VAL_PWR_MGMT_2_LP_WAKE_CTRL_40HZ    = 0xC0
    
    VAL_PWR_MGMT_2_STBY_XA = 0x20
    VAL_PWR_MGMT_2_STBY_YA = 0x10
    VAL_PWR_MGMT_2_STBY_ZA = 0x08
    
    VAL_PWR_MGMT_2_STBY_XG = 0x04
    VAL_PWR_MGMT_2_STBY_YG = 0x02
    VAL_PWR_MGMT_2_STBY_ZG = 0x01
    
    #INT enable values
    VAL_INT_ENABLE_DISABLED = 0x00      #Disabled INT
    VAL_INT_ENABLE_MOTION   = 0x40      #Motion detection
    VAL_INT_ENABLE_DATA_RDY = 0x01      #Data Ready ( Gyro data do not have data in cycle mode )

    _address = None
    _bus = None
    _gravityFactor = None;

    def __init__(self, address, busId = 1, gravityFactor = GRAVITIY_EARTH ):
        """!
        \~english
        Initialize the MPU6050 object instance

        @param address        MPU6050 I2C Address (default is: 0x68)
        @param busId          I2C Bus ID of Raspberry Pi. RPi-Spark pHAT default is 1
        @param gravityFactor  Default is GRAVITIY_EARTH
        
        \~chinese
        初始化 MPU6050 实例

        @param address        MPU6050 I2C 地址（默认为：0x68）
        @param busId          树梅派( Raspberry Pi ) I2C 总线 ID。 RPi-Spark pHAT 默认值为 1
        @param gravityFactor  默认 GRAVITIY_EARTH ( 地球重力加速度 )

        \~ \n
        @see DEF_MPU6050_ADDRESS
        @see GRAVITIY_EARTH
        """
        self._address = address
        self._bus = smbus.SMBus( busId )
        self._gravityFactor = gravityFactor
        self.setAccelRange( self.ACCEL_RANGE_2G )

    def _readByte(self, regAddr):
        return self._bus.read_byte_data(self._address, regAddr)

    def _readWord(self, regAddr):
        high = self._readByte(regAddr)
        low = self._readByte(regAddr+1)
        rawVal = ( high << 8 ) | low;
        if (rawVal >= 0x8000):
            return -((65535 - rawVal) + 1)
        else:
            return rawVal

    def _writeByte(self, regAddr, regValue):
        self._bus.write_byte_data(self._address, regAddr, regValue)

    def _sendCmd(self, cmd, value, firstClear = True):
        if firstClear:
            self._writeByte(cmd, 0x00)
        self._writeByte(cmd, value)

    def reset(self):
        """!
        \~english Reset MPU all registers
        \~chinese 复位全部寄存器
        """
        self._writeByte(self.REG_PWR_MGMT_1, self.VAL_PWR_MGMT_1_RESET)
 
    def open(self):
        """!
        \~english Trun on device and with all sensors at same time
        \~chinese 开启全部传感器
        """
        self._sendCmd(self.REG_PWR_MGMT_1, 0x00)
        self._sendCmd(self.REG_PWR_MGMT_2, 0x00)

    def openWith(self, accel = True, gyro = True, temp = True, cycle = False, cycleFreq = 0x00):
        """!
        Trun on device with configurable sensors into wake up mode
        
        @param accel: True - Enable accelerometer
        @param gyro: True - Enable gyroscope
        @param temp: True - Enable Thermometer
        @param cycle: True - Enable cycle wake-up mode
        @param cycleFreq: Cycle wake-up frequency, this value can be chosen:

          @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_1_25HZ is default
          @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_5HZ
          @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_20HZ
          @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_40HZ
        """
        val_pwr_2 = 0x00
        #disable Accel sensor
        if accel == False:
            val_pwr_2 = val_pwr_2 | self.VAL_PWR_MGMT_2_STBY_XA
            val_pwr_2 = val_pwr_2 | self.VAL_PWR_MGMT_2_STBY_YA
            val_pwr_2 = val_pwr_2 | self.VAL_PWR_MGMT_2_STBY_ZA

        #disable Gyro sensor
        if gyro == False:
            val_pwr_2 = val_pwr_2 | self.VAL_PWR_MGMT_2_STBY_XG
            val_pwr_2 = val_pwr_2 | self.VAL_PWR_MGMT_2_STBY_YG
            val_pwr_2 = val_pwr_2 | self.VAL_PWR_MGMT_2_STBY_ZG

        val_pwr_1 = 0x00
        #disable temp sensor
        if temp == False:
            val_pwr_1 = val_pwr_1 | self.VAL_PWR_MGMT_1_OFF_TEMP

        #cycle mode
        if cycle == True:
            val_pwr_1 = val_pwr_1 | self.VAL_PWR_MGMT_1_ON_CYCLE
            val_pwr_2 = val_pwr_2 | cycleFreq

        self._sendCmd( self.REG_PWR_MGMT_2, val_pwr_2 )
        self._sendCmd( self.REG_PWR_MGMT_1, val_pwr_1 )

    def openOnlyAccel(self, cycleFreq = 0x00 ):
        """!
        Trun on device into Accelerometer Only Low Power Mode
        @param cycleFreq can be choise: 
            @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_1_25HZ is default
            @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_5HZ
            @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_20HZ
            @see VAL_PWR_MGMT_2_LP_WAKE_CTRL_40HZ
        """
        self.openWith(accel = True, gyro = False, temp = False, cycle = True, cycleFreq = cycleFreq)

    def sleep(self):
        """!
            \~english Set device into sleep mode
            \~chinese MPU6050 进入睡眠模式，低功耗
        """
        self._sendCmd(self.REG_PWR_MGMT_1, self.VAL_PWR_MGMT_1_SLEEP)

    def setMotionInt(self, motDHPF = 0x01, motTHR = 0x14, motDUR = 0x30, motDeteDec = 0x15 ):
        """!
        Set to enable Motion Detection Interrupt

        @param motDHPF Set the Digital High Pass Filter. Default is 0x01 (5Hz)
        @param motTHR  Desired motion threshold. Default is 20 (0x14)
        @param motDUR  Desired motion duration. Default is 48ms (0x30)
        @param motDeteDec Motion detection decrement. Default is 21 (0x15)
        
        @note <b>motDHPF</b> should be one of the following values:<br>
            0x00: RESET,<br>
            0x01: 5Hz,<br>
            0x02: 2.5Hz,<br>
            0x03: 1.25Hz,<br>
            0x04: 0.63Hz,<br>
            0x07: HOLD<br>
        """
        #After power on (0x00 to register (decimal) 107), the Motion Detection Interrupt can be enabled as follows:
        #self._sendCmd( self.REG_PWR_MGMT_1, 0x00 )
        #(optionally?) Reset all internal signal paths in the MPU-6050 by writing 0x07 to register 0x68;
        self._sendCmd( self.REG_SIGNAL_PATH_RESET, 0x07 )
        #write register 0x37 to select how to use the interrupt pin. 
        #For an active high, push-pull signal that stays until register 
        #(decimal) 58 is read, write 0x20 (need read to clear INT state) or 0x00 (auto clear INT state).
        self._sendCmd( self.REG_INT_PIN_CFG, 0x00 )
        
        orgAccelConf = self._readByte(self.REG_ACCEL_CONFIG)
        newAccelConf = ( (orgAccelConf | 0xE7) ^ 0xE7 ) | motDHPF
        # Write register 28 (==0x1C) to set the Digital High Pass Filter, 
        # bits 3:0. For example set it to 0x01 for 5Hz. 
        # (These 3 bits are grey in the data sheet, but they are used! 
        # Leaving them 0 means the filter always outputs 0.)
        #
        # 0x00: RESET,
        # 0x01: 5Hz,
        # 0x02: 2.5Hz,
        # 0x03: 1.25Hz,
        # 0x04: 0.63Hz,
        # 0x07: hold
        #
        # 高通滤波器灵敏度调节
        #
        self._sendCmd( self.REG_ACCEL_CONFIG, newAccelConf )
        #Write the desired Motion threshold to register 0x1F (For example, write decimal 20).
        self._sendCmd( self.REG_MOTION_DET, motTHR )
        #To register 0x20 (hex), write the desired motion duration, for example 40ms (0x28).
        # 灵敏度调节
        self._sendCmd( self.REG_MOTION_DET_DUR, motDUR )
        
        #to register 0x69, write the motion detection decrement and 
        #a few other settings (for example write 0x15 to set both 
        #free-fall and motion decrements to 1 and accelerome0x00ter 
        #start-up delay to 5ms total by adding 1ms. )
        self._sendCmd( self.REG_MOTION_DET_CTRL, motDeteDec )

        #write register 0x38, bit 6 (0x40), to enable motion detection interrupt.
        self._sendCmd( self.REG_INT_ENABLE, self.VAL_INT_ENABLE_MOTION )

    def setDataRdyInt(self, int_cfg = 0x20 ):
        """!
        \~english 
        Set to enabled Data Ready Interrupt
        int_cfg : Register 55( 0x37 ) – INT Pin / Bypass Enable Configuration, page 26
        
        \~chinese 
        启用数据就绪中断
        @param int_cfg: 寄存器 55( 0x37 ) – INT Pin / Bypass Enable Configuration, page 26
        """
        self._sendCmd( self.REG_INT_PIN_CFG, int_cfg )
        self._sendCmd( self.REG_INT_ENABLE, self.VAL_INT_ENABLE_DATA_RDY)

    def disableInt(self):
        """!
            Trun Off Interrupt
        """
        self._sendCmd( self.REG_INT_ENABLE, self.VAL_INT_ENABLE_DISABLED )

    def getIntDataRdy(self):
        """!
        \~english
        Get data ready interrupt status (data ready, motion detection, etc.) 
        After calling this method the data ready interrupt flag will be reset, 
        waiting for the next data ready interrupt

        MPU6050 Register Map and Descriptions revision 4.2, page 28 
        Register 58(0x3A) – Interrupt Status
        \~chinese
        读取数据就绪中断状态（数据就绪，运动检测等），调用该方法后数据就绪中断标志位将复位，等待下一次数据就绪中断
        MPU6050寄存器映射和描述修订4.2，第28页
        寄存器58（0x3A）- 中断状态
        """
        return self._readByte( self.REG_INT_STATUS )

    def getTemp(self):
        """!
        \~english
        Reads the temperature from the onboard temperature sensor of the MPU-6050.
        @returns a float value, the temperature in degrees Celcius.
        @note MPU6050 Register Map and Descriptions revision 4.2, page 30
        \~chinese
        MPU-6050 的板载温度传感器读取温度
        @returns 浮点值，温度以摄氏度为单位
        @note MPU6050寄存器映射和描述修订4.2，第30页
        """
        rawTemp = self._readWord( self.REG_TEMP_OUT_H )
        return (rawTemp + 12412.0) / 340.0
        # Get the actual temperature using the formule given in the
        # MPU-6050 Register Map and Descriptions revision 4.2, page 30
        # Temperature in degrees C = (TEMP_OUT Register Value as a signed quantity)/340 + 36.53
#         return ( rawTemp / 340.0 ) + 36.53

    def setAccelRange(self, accelRange):
        """!
        Set range of accelerometer.
        @param accelRange: range of accelerometer.
          It should be one of the following values:
              @see ACCEL_RANGE_2G
              @see ACCEL_RANGE_4G
              @see ACCEL_RANGE_8G
              @see ACCEL_RANGE_16G
        @note Using a pre-defined range is advised.
        """
        self._sendCmd( self.REG_ACCEL_CONFIG, accelRange )

    def readAccelRange( self ):
        """!
        Reads the range of accelerometer setup.
        
        @return an int value.
          It should be one of the following values:
              @see ACCEL_RANGE_2G
              @see ACCEL_RANGE_4G
              @see ACCEL_RANGE_8G
              @see ACCEL_RANGE_16G
        """
        raw_data = self._readByte(self.REG_ACCEL_CONFIG)
        raw_data = (raw_data | 0xE7) ^ 0xE7
        return raw_data

    def getAccelData( self,  raw = False ):
        """!
        Gets and returns the X, Y and Z values from the accelerometer.

        @param raw If raw is True, it will return the data in m/s^2,<br> If raw is False, it will return the data in g
        @return a dictionary with the measurement results or Boolean.
            @retval {...} data in m/s^2 if raw is True.
            @retval {...} data in g if raw is False.
            @retval False means 'Unkown accel range', that you need to check the "accel range" configuration
        @note Result data format: {"x":0.45634,"y":0.2124,"z":1.334}
        """
        x = self._readWord(self.REG_ACCEL_XOUT_H)
        y = self._readWord(self.REG_ACCEL_YOUT_H)
        z = self._readWord(self.REG_ACCEL_ZOUT_H)

        accel_scale_modifier = None
        accel_range = self.readAccelRange()

        if accel_range == self.ACCEL_RANGE_2G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_2G
        elif accel_range == self.ACCEL_RANGE_4G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_4G
        elif accel_range == self.ACCEL_RANGE_8G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_8G
        elif accel_range == self.ACCEL_RANGE_16G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_16G
        else:
            print( "ERROR: Unkown accel range!" )
            return False            
            #accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_2G

        x = x / accel_scale_modifier
        y = y / accel_scale_modifier
        z = z / accel_scale_modifier

        if raw == True:
            return { 'x': x, 'y': y, 'z': z }
        elif raw == False:
            return { 'x': x * self._gravityFactor, 'y': y * self._gravityFactor, 'z': z * self._gravityFactor }

    def setGyroRange(self, gyroRange):
        """!
        Set range of gyroscope.
        @param gyroRange: range of gyroscope.
          It should be one of the following values:
              @see GYRO_RANGE_250DEG
              @see GYRO_RANGE_500DEG
              @see GYRO_RANGE_1KDEG
              @see GYRO_RANGE_2KDEG 

        @note Using a pre-defined range is advised.
        """
        self._sendCmd( self.REG_GYRO_CONFIG, gyroRange )

    def readGyroRange( self ):
        """!
        Read range of gyroscope.

        @return an int value. It should be one of the following values (GYRO_RANGE_250DEG)

        @see GYRO_RANGE_250DEG
        @see GYRO_RANGE_500DEG
        @see GYRO_RANGE_1KDEG
        @see GYRO_RANGE_2KDEG
        """
        raw_data = self._readByte( self.REG_GYRO_CONFIG )
        raw_data = (raw_data | 0xE7) ^ 0xE7
        return raw_data

    def getGyroData(self):
        """!
        Gets and returns the X, Y and Z values from the gyroscope

        @return a dictionary with the measurement results or Boolean.
            @retval {...} a dictionary data.
            @retval False means 'Unkown gyroscope range', that you need to check the "gyroscope range" configuration
        @note Result data format: {"x":0.45634,"y":0.2124,"z":1.334}
        """
        x = self._readWord(self.REG_GYRO_XOUT_H)
        y = self._readWord(self.REG_GYRO_YOUT_H)
        z = self._readWord(self.REG_GYRO_ZOUT_H)

        gyro_scale_modifier = None
        gyro_range = self.readGyroRange()

        if gyro_range == self.GYRO_RANGE_250DEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_250DEG
        elif gyro_range == self.GYRO_RANGE_500DEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_500DEG
        elif gyro_range == self.GYRO_RANGE_1KDEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_1KDEG
        elif gyro_range == self.GYRO_RANGE_2KDEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_2KDEG
        else:
            print("ERROR: Unkown gyroscope range!")
            return False
            #gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_250DEG

        x = x / gyro_scale_modifier
        y = y / gyro_scale_modifier
        z = z / gyro_scale_modifier
        return {'x': x, 'y': y, 'z': z}


    def getAllData(self, temp = True, accel = True, gyro = True):
        """!
        Get all the available data.

        @param temp: True - Allow to return Temperature data
        @param accel: True - Allow to return Accelerometer data
        @param gyro: True - Allow to return Gyroscope data

        @return a dictionary data
            @retval {} Did not read any data
            @retval {"temp":32.3,"accel":{"x":0.45634,"y":0.2124,"z":1.334},"gyro":{"x":0.45634,"y":0.2124,"z":1.334}} Returned all data
        """
        allData = {}
        if temp:
            allData["temp"] = self.getTemp()

        if accel:
            allData["accel"] = self.getAccelData( raw = False )

        if gyro:
            allData["gyro"] = self.getGyroData()

        return allData

#
# This a simple test
# if __name__ == "__main__":
#     mpu = mpu6050( DEF_MPU6050_ADDRESS )
#     print("REG_PWR_MGMT_1: {:#4x}".format(mpu._readByte(mpu.REG_PWR_MGMT_1)))
# #     mpu.reset()
#     mpu.setAccelRange( mpu.ACCEL_RANGE_2G )
#     mpu.open()
#     print("REG_PWR_MGMT_1: {:#4x}".format(mpu._readByte(mpu.REG_PWR_MGMT_1)))
#     print("Accel Range: {:#4x} -- {:d}g".format(mpu.readAccelRange(), mpu.readAccelRange()))
#     print("Gryo Range: {:#4x}".format(mpu.readGyroRange()))
# 
#     print (mpu.getAccelData( False ))
#     print ("---------------------")
# 
#     allData = mpu.getAllData()
#     print (allData)
#     mpu.sleep()


# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
#    RPi-Spark PILScreen
#    by Kunpeng Zhang
#    v1.0.0
#    
#    Use PIL as cache for draw canvas
#    使用 PIL 作为绘制缓存的显示屏实现
#
import types

import PIL
from PIL import ImageFont
from PIL import Image
from PIL import ImageDraw

from .SScreen import SScreenBase
from .SScreen import SSRect

SCR_DEF_FRONT = ImageFont.load_default()

class SSPILScreen( SScreenBase ):
    """!
    \~english This class work with PIL Image Lib.
    \~chinese 使用 PIL 库实现的 SScreen 类
    """

    def _catchCurrentViewContent(self):
        """!
        \~english
        Catch the current view content
        @return: a PIL Image
        @note 
            Automatically converts the cache color mode and at the 
            same time rotates the captured image data according to 
            the screen angle

        \~chinese
        从缓存中抓取当前视图大小的数据
        @return: PIL Image 对象
        @note 自动转换缓存色彩模式，同时根据屏幕角度设定旋转所抓取的图像数据
        """
        viewContent = None
        if self._buffer_color_mode != self._display_color_mode:
            viewContent = self._buffer.crop( self.View.rectToArray() ) .convert( self._display_color_mode )
        else:
            viewContent = self._buffer.crop( self.View.rectToArray() )

        # Rotate for display direction
        if self._display_direction == 0:
            return viewContent
        else:
            return viewContent.rotate( angle = self._display_direction, expand=True )

    def _initBuffer(self, bufferColorMode, bufferSize):
        """!
        \~english
        Initialize the buffer object instance, use PIL Image as for buffer
        @param bufferColorMode: "RGB" or "1"
        @param bufferSize: (width, height)
        \~chinese
        初始化缓冲区对象实例，使用PIL Image作为缓冲区
        @param bufferColorMode: 色彩模式, 取值： "RGB" 或 "1"
        @param bufferSize: 缓存大小 (width, height)，例如： (128, 64)
        """
        # super(SSScreenBase)._initBuffer(bufferColorMode, bufferSize)
        self._buffer_color_mode = bufferColorMode

        #create screen image buffer and canvas
        if bufferSize==None:
            self._buffer = Image.new( bufferColorMode , self._display_size )
        else:
            self._buffer = Image.new( bufferColorMode , bufferSize )        
        self.Canvas = ImageDraw.Draw( self._buffer )

        #creare screen view
        self.View = SSRect( 0, 0, self._display_size[0], self._display_size[1] )

    def getBufferSize(self):
        return self._buffer.size

#     def refresh(self):
#         """Update current view content to display
#         """
#         pass

    def clearCanvas(self, fillColor = 0 ):
        """!
        \~engliash
        Clear up canvas and fill color at same time
        @param fillColor:  a color value
        @note 
            The fillColor value range depends on the setting of _buffer_color_mode.
            * If it is SS_COLOR_MODE_MONO ("1") monochrome mode, it can only select 0: black and 1: white
            * If it is SS_COLOR_MODE_RGB ("RGB") color mode, RGB color values can be used
        \~chinese
        清除画布并同时填充颜色
        @param fillColor: 颜色值
        @note 
            fillColor 取值范围取决于 _buffer_color_mode 的设定。
            * 如果是 SS_COLOR_MODE_MONO ("1") 单色模式，只能选择 0:黑色 和 1:白色 
            * 如果是 SS_COLOR_MODE_RGB ("RGB") 彩色模式，可以使用 RGB 色彩值
        """
        self.Canvas.rectangle((0, 0, self._display_size[0], self._display_size[1]), outline=0, fill=fillColor)
    
    def clearView(self, fillColor = 0 ):
        """!
        \~english
        Clear up canvas with view size
        @param fillColor: a color value
        @note 
            The fillColor value range depends on the setting of _buffer_color_mode.
            * If it is SS_COLOR_MODE_MONO ("1") monochrome mode, it can only select 0: black and 1: white
            * If it is SS_COLOR_MODE_RGB ("RGB") color mode, RGB color values can be used

        \~chinese
        清除画布中当前视图大小的区域同时填充颜色
        @param fillColor: 颜色值
        @note 
            fillColor 取值范围取决于 _buffer_color_mode 的设定。
            * 如果是 SS_COLOR_MODE_MONO ("1") 单色模式，只能选择 0:黑色 和 1:白色 
            * 如果是 SS_COLOR_MODE_RGB ("RGB") 彩色模式，可以使用 RGB 色彩值
        """
        self.Canvas.rectangle(self.View.rectToArray(), outline=0, fill=fillColor)

    def clear(self):
        """!
        \~english
        Clear display and screen's canvas
        \~chinese
        同时清除显示屏和画布
        """
        self.clearCanvas()
        self.Display.clear()
        pass

    def redefineBuffer(self, newBuffer ):
        """!
        \~english 
        Redefine frame of Screen
        @param newFrame: a new fram data
        @note
            newFrame can be:
            * PIL Image
            * PIL ImageFile
            * Dictionary, eg. { "size":(width, height), "color_mode":"1" } or { "size":(width, height), "color_mode":"RGB" }

        \~chinese
        重新定义缓存数据
        @param newFrame: 新缓存数据 \n
            newFrame 可以为下面值:
            * PIL Image
            * PIL ImageFile
            * 字典, eg. { "size":(width, height), "color_mode":"1" } or { "size":(width, height), "color_mode":"RGB" } 
        """
        # Redefine Frame from an image object
        if type(self._buffer) == type(newBuffer):
            self._buffer = newBuffer
            self.Canvas = ImageDraw.Draw( self._buffer )
#             self.View.resize(newBuffer.width, newBuffer.height)
            return True

        # Redefine Frame from an <PIL.ImageFile.ImageFile>
        if type(newBuffer).__name__.find(PIL.ImageFile.ImageFile.__name__) != -1:        
            self._buffer = self._buffer.resize((newBuffer.width, newBuffer.height))
            self._buffer.paste( newBuffer, (0,0))
#             self.View.resize(newBuffer.width, newBuffer.height)
            return True

        # Recreated a new frame from dict of frame
        if isinstance(newBuffer, dict):
            self._buffer = Image.new( newBuffer["color_mode"] , newBuffer["size"] )
            self.Canvas = ImageDraw.Draw( self._buffer )
            return True
        pass

    def write(self, text="", xy=(0,0), align="left", font=None, fontName=None, fontSize = 10, fill = 1, spacing = 0, screenCenter = False):
        """!
        \~english 
        Print one line text or multi-line text on the screen
        @param text: Text to be drawn. eg. "Hello World!" or "Hello/nWorld!"
        @param xy: Top left corner of the text. defaule: (0,0)
        @param align: "left", "center" or "right". defaule: "left"
        @param fontName: Font of text. defaule: None (default font) 
        @param fontSize: Font size. default: 10
        @param fill: Color to use for the text. default: 1 (white)
        @param spacing: The number of pixels between lines. default: 0
        @param screenCenter： Center the text on the screen. default: False

        \~chinese
        在屏幕上打印一行文字或者多行文字
        @param text: \n
        @param xy:\n
        @param align:\n
        @param fontName:\n
        @param fontSize:\n
        @param fill:\n
        @param spacing:\n
        @param screenCenter:\n
        
        @param text: 要输出的文字，可以单行也可以多行。例如： "Hello World!"  或 "Hello/nWorld!"
        @param xy: 文字输出的坐标点。默认： (0,0)
        @param align: 多行文字对齐方式，可选： "left", "center" 或 "right". 默认： "left"
        @param fontName: 字体名。默认：None（使用系统默认的字体） 
        @param fontSize: 字体大小。默认：10
        @param fill: 文字颜色。默认： 1 （白色）
        @param spacing: 行间距。默认：0
        @param screenCenter： 让文本位于屏幕居中。
        """
        tx = xy[0]
        
        try:
            dwFont = font if font != None else SCR_DEF_FRONT if fontName==None else ImageFont.truetype(fontName, fontSize)
        except:
            dwFont = ImageFont.load_default()

        try:
            if screenCenter == True:
                (fw, fh) = self.Canvas.multiline_textsize( text, font )
                tx = xy[0] + (self._display_size[0]-fw)/2
            self.Canvas.multiline_text( (tx, xy[1]) , text, font = dwFont, align=align, fill=fill, spacing=spacing)
        except:
            print("ERROR: canvas write error")



# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
# RPi-Spark pHAT -- Key Buttons
# Keyboard include Joystick buttons and Action buttons, 
# keyboard use BCM mode, there are keyboard layout:
# 
#              [JOY UP]
# [JOY LEFT]   [JOY OK]   [JOY RIGHT]             [ACT_A]  [ACT_B]
#              [JOY DOWN]
#
# @author Kunpeng Zhang
# v1.0    2018.3.8
#

import RPi.GPIO as GPIO

# Action Buttons    BCM_IO_NUM
# BUTTON_ACT_A        = 22
# BUTTON_ACT_B        = 23

# Joy Buttons       BCM_IO_NUM
# BUTTON_JOY_LEFT     = 26
# BUTTON_JOY_RIGHT    = 27
# BUTTON_JOY_UP       = 5
# BUTTON_JOY_DOWN     = 6
# BUTTON_JOY_OK       = 24

# UNIT: ms
DEF_BOUNCE_TIME_SHORT_MON   = 10
DEF_BOUNCE_TIME_SHORT       = 50
DEF_BOUNCE_TIME_NORMAL      = 100
DEF_BOUNCE_TIME_LONG        = 200

#SparkKeyboard
class RPiKeyButtons :
    """!
    \~english This RPi-Spark pHAT Key Buttons Drive
    \~chinese 树梅派火花(RPi-Spark pHAT) 按键驱动
    """
    def __init__(self):
        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)
        pass

    def setKeyButton( self, btnId, keyCallback, bounceTime = DEF_BOUNCE_TIME_NORMAL, pullUpDown = GPIO.PUD_UP, event = GPIO.BOTH ):
        """!
        \~english
        Set key button event
        @param btnId: Key button pin number in BCM
        @param keyCallback: A interrupt <b>callback_function</b> or <b>None</b>. <br>
               If set to <b>None</b> means keybutton work in query mode<br>
               then uses RPiKeyButtons#readKeyButton for get keybutton status
        @param bounceTime: Default set to DEF_BOUNCE_TIME_NORMAL
        @param pullUpDown: Default set to GPIO.PUD_UP
        @param event: Default set to GPIO.BOTH. it can be: { GPIO.RISING | GPIO.FALLING | GPIO.BOTH }

        \~chinese
        设置按键事件
        @param btnId: 按键IO号(in BCM mode)
        @param keyCallback: 按键中断回调函数 <b>callback_function</b> 或 <b>None</b><br>
                如果设置为<b>None</b>，则表示按键工作在查询模式<br>
                然后使用 RPiKeyButtons#readKeyButton 获取 keybutton 状态
        @param bounceTime: 默认 DEF_BOUNCE_TIME_NORMAL
        @param pullUpDown: 默认 GPIO.PUD_UP
        @param event: 默认 GPIO.BOTH 它可以是: { GPIO.RISING | GPIO.FALLING | GPIO.BOTH }

        \~ \n
        @see DEF_BOUNCE_TIME_SHORT_MON (10ms)
        @see DEF_BOUNCE_TIME_SHORT (50ms)
        @see DEF_BOUNCE_TIME_NORMAL (100ms)
        @see DEF_BOUNCE_TIME_LONG (200ms)

        @note
            * setKeyButton(btnId = 12, keyCallback = None )
            * setKeyButton(btnId = 14, keyCallback = aKeyCallbackFun ) <br>
                <pre>
                \# a simple callback function
                def aKeyCallbackFun(channel):
                    print(channel)
                    pass
                </pre>
        """
        GPIO.setup( btnId, GPIO.IN, pull_up_down=pullUpDown)
        # The keyCallback is None means setting keybutton in query mode, 
        # then uses readKeyButton for get keybutton status
        # event can be { RISING, FALLING, BOTH }
        if keyCallback != None:
            try:
                GPIO.add_event_detect( btnId, event, callback=keyCallback, bouncetime=bounceTime )
            except:
                pass
        pass

    def setKeyButtonCallback(self, btnId, keyCallback):
        """!
        \~english
        Set key button event callback
        @param btnId: button id number of GPIO. eg. 13 or 15 etc.
        @param keyCallback: a callback function. eg.\n
        \~chinese
        设置按键事件回调
        @param btnId: 按键IO号(in BCM mode) 例如: 13 or 15 etc.
        @param keyCallback: 回调函数 例如:\n
        \~
        <pre>
        def keyCallback(channel):
           print(channel)
           pass
        </pre>
        """
        GPIO.add_event_callback( btnId, keyCallback )

    def removeKeyButtonEvent(self, buttons= [] ):
        """!
        \~english
        Remove key button event callbacks
        @param buttons: an array of button Ids. eg. [ 12,13,15, ...]

        \~chinese
        移除按键事件回调
        @param buttons: 按钮ID数组。 例如: [12,13,15，...]
        """
        for i in range( 0, len(buttons)-1 ):
            GPIO.remove_event_detect( buttons[i] )

    def readKeyButton( self, btnId ):
        """!
        \~english
        Read an button IO status
        @return an boolean value
        \~chinese
        读取按钮IO状态
        @return 布尔值
        """
        return GPIO.input(btnId) if btnId != None else False

    def configKeyButtons( self, enableButtons = [], bounceTime = DEF_BOUNCE_TIME_NORMAL, pullUpDown = GPIO.PUD_UP, event = GPIO.BOTH ):
        """!
        \~english
        Config multi key buttons IO and event on same time

        @param enableButtons: an array of key button configs. eg. <br>
                [{ "id":BUTTON_ACT_A, "callback": aCallbackFun }, ... ]
        @param bounceTime: Default set to DEF_BOUNCE_TIME_NORMAL
        @param pullUpDown: Default set to GPIO.PUD_UP
        @param event: Default set to GPIO.BOTH. it can be: { GPIO.RISING | GPIO.FALLING | GPIO.BOTH }

        \~chinese
        同时配置多个按键IO和事件

        @param enableButtons: 组按键配置 例如: <br>
                [{ "id":BUTTON_ACT_A, "callback": aCallbackFun }, ... ]
        @param bounceTime: 默认 DEF_BOUNCE_TIME_NORMAL
        @param pullUpDown: 默认 GPIO.PUD_UP
        @param event: 默认 GPIO.BOTH 它可以是: { GPIO.RISING | GPIO.FALLING | GPIO.BOTH }

        \~ \n
        @see DEF_BOUNCE_TIME_SHORT_MON (10ms)
        @see DEF_BOUNCE_TIME_SHORT (50ms)
        @see DEF_BOUNCE_TIME_NORMAL (100ms)
        @see DEF_BOUNCE_TIME_LONG (200ms)
        """
        for key in enableButtons:
            self.setKeyButton( key["id"], key["callback"], bounceTime, pullUpDown, event )
        pass



# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
#    RPi-Spark ScreenBase
#    by Kunpeng Zhang
#    v1.0.0
#

# Display color mode

## Mono color mode
SS_COLOR_MODE_MONO  = "1"
## RGB color mode
SS_COLOR_MODE_RGB   = "RGB"

class SSPoint:
    """!
    \~english A point object
    \~chinese 点对象
    """
    x = 0
    y = 0

class SSRect:
    """!
    \~english A rectangles object
    \~chinese 矩形对象
    """
    x       = 0
    y       = 0
    height  = 0
    width   = 0

    def __init__(self, x, y, width, height):
        """!
        \~english Initialize the rectangles object instance
        \~chinese 初始化矩形实例
        """
        self.x = x
        self.y = y
        self.height = height
        self.width = width

    def resize(self, newWidth = 0, newHeight = 0):
        """!
        \~english 
        Resize width and height of rectangles
        @param newWidth: new width value
        @param newHeight: new height value
        \~chinese 
        重新设定矩形高宽
        @param newWidth: 新宽度
        @param newHeight: 新高度
        """
        self.height = newHeight
        self.width = newWidth

    def adjuestSize(self, offsetWidth = 0, offsetHeight = 0):
        """!
        \~english 
        Adjuest width and height of rectangles
        @param offsetWidth: adjust the width. Negative numbers are smaller, Positive number is increased
        @param offsetHeight: adjust the height. Negative numbers are smaller, Positive number is increased
        
        @note The negative numbers are smaller, positive number is increased，0 remains unchanged.
        
        \~chinese
        调整矩形高宽数据
        @param offsetWidth: 调整宽度。 负数较小，正数增加
        @param offsetHeight: 调整高度。 负数较小，正数增加

        @note 负数较小，正数增加，0保持不变。
        """
        self.height += offsetHeight
        self.width += offsetWidth

    def moveTo(self, newX=0, newY=0):
        """!
        \~english 
        Move vertex of rectangles to new point (x,y)
        @param newX: Coordinated X value
        @param newY: Coordinated Y value
        \~chinese
        移动矩形到新坐标点 (x,y)
        @param newX: 坐标 X
        @param newY: 坐标 Y
        """
        self.x = newX
        self.y = newY

    def moveOffset(self, offsetX=0, offsetY=0):
        """!
        \~english 
        Offset vertex of rectangles to new point (x,y)
        @param offsetX: offset X value
        @param offsetY: offset Y value
        @note The negative numbers are left or up move , positive number is right or down move，0 remains unchanged.
        \~chinese
        平移矩形指定的距离 (x,y)
        @param offsetX: 平移 X
        @param offsetY: 平移 Y
        @note 负数是左移( X )或上移( Y )，正数是右移( X )或下移( Y )，0 保持不变。
        """
        self.x += offsetX
        self.y += offsetY

    def swapWH(self):
        """!
        \~english Swap width and height of rectangles
        \~chinese 交换矩形高宽边数据
        """
        width = self.width
        self.width = self.height
        self.height = width

    def rectToArray(self, swapWH = False):
        """!
        \~english
        Rectangles converted to array of coordinates
        @return: an array of rect points. eg. (x1,y1,x2,y2)

        \~chinese
        矩形数据转换为矩形坐标数组
        @return: 矩形座标数组, 例如: ( x1,y1,x2,y2 )
        """
        if swapWH == False:
            return [self.x, self.y, self.x + self.width, self.y + self.height]
        else:
            return [self.x, self.y, self.x + self.height, self.y + self.width]

class SScreenBase:
    """!
    \~english
    SScreenBase is a hardware abstraction of the screen，
    it contain a Display, a Canvas and a View instances
    You need to create a subclass from inherit it and 
    use the new subclass implement initialization and 
    operations of screen.

    \~chinese
    SScreenBase是屏幕的硬件抽象，它包含一个 Display，一个 Canvas 和一个 View 实例
    你需要通过继承和创建一个子类, 使用新的子类实现初始化和屏幕的操作。
    """
    ##
    # \~english a display hardware instance
    # \~chinese 显示屏芯片硬件实例
    Display = None
    ##
    # \~english a canvas instance. you can draw something on it, then call SScreenBase#refresh to update to display
    # \~chinese 画布实例，可以绘制图像文字，然后通过 SScreenBase#refresh 更新到显示屏
    Canvas = None
    ##
    # \~english a SSRect instance. 
    # \~chinese SSRect 实例.
    View = None

    # Display direction can choos in 0, 90, 180, 270
    _display_direction = None
    # Display's size, it init at intance when __init__ (x,y), and it keep same direction whit "_display_direction"
    # If you need to access the physical size, use the “Display” object
    _display_size = None #(0,0)
    
    ##
    # display color mode, it is readonly, can not be change. 
    # it can be chosen: SS_COLOR_MODE_MONO ("1"): greyscale or SS_COLOR_MODE_RGB ("RGB"): image
    _display_color_mode = None #SS_COLOR_MODE_MONO
    
    ##
    # buffer color mode, it is can be change using SScreenBase#changeBufferColorMode,
    # and it can diffent from _display_color_mode,
    # its value could be: SS_COLOR_MODE_MONO ("1"): greyscale or SS_COLOR_MODE_RGB ("RGB"): color
    _buffer_color_mode = None #SS_COLOR_MODE_MONO

    ##
    # Screen buffer, its size can diffent from display size.
    _buffer = None

    def __init__(self, display, bufferColorMode, bufferSize=None, displayDirection=0 ):
        raise NotImplementedError

    def _initDisplay(self, display, displayDirection, displaySize=(0,0)):
        self._display_direction = displayDirection

        if self._needSwapWH(0, displayDirection):
            self._display_size = ( displaySize[1], displaySize[0] )
        else:
            self._display_size = displaySize

        self.Display = display

    def _initBuffer(self, bufferColorMode, bufferSize):
        self._buffer_color_mode = bufferColorMode

    def _needSwapWH(self, oldDirection, newDirection ):
        """!
        \~english
        return screen direction status
        @return Boolean
        @note No need to rotate if the screen orientation is 0 degrees and 180 degrees

        \~chinese
        返回屏幕方向状态
        @return 布尔值
        @note 如果屏幕方向是0度和180度就不需要旋转
        """
        if abs(newDirection - oldDirection) == 0: return False
        if abs(newDirection - oldDirection) % 180 == 0: return False
        if abs(newDirection - oldDirection) % 90 == 0: return True
        return False

    def _checkBufferColorMode(self, bufferColorMode):
        """!
        \~english
        Check buffer color mode
        @param bufferColorMode: 

        \~chinese
        检测缓存色彩模式
        @param bufferColorMode: 
        """
        if bufferColorMode != SS_COLOR_MODE_RGB and bufferColorMode != SS_COLOR_MODE_MONO:
            raise ValueError("Incorrect bufferColorMode mode, this value just can be chosen: \"RGB\" or \"1\" ")

    def refresh(self):
        """!
        \~english Update current view content to display
        \~chinese 更新当前 View 内容到显示屏
        """
        raise NotImplementedError

    def clear(self):
        """!
        \~english Clear screen
        \~chinese 清屏
        """
        raise NotImplementedError

    def rotateDirection(self, displayDirection):
        """!
        \~english rotate screen direction
        @param displayDirection: Screen Direction. value can be chosen: 0, 90, 180, 270
        \~chinese 旋转显示屏方向
        @param displayDirection: 显示屏方向。可选值： 0, 90, 180, 270
        
        \~
        @note
        \~english after rotate the View resize to screen size 
        \~chinese 改变方向后，默认的 View 大小会更新为当前 Screen 的大小
        \~\n
        """
        if self._needSwapWH(self._display_direction, displayDirection):
            self._display_size = ( self._display_size[1], self._display_size[0] )
            if self.redefineBuffer( { "size":self._display_size, "color_mode":self._buffer_color_mode } ):
                self.View.resize(self._display_size[0], self._display_size[1])
        self._display_direction = displayDirection

    def redefineBuffer(self, newFrame ):
        """!
        \~english 
        Redefine frame of Screen
        @param newFrame: an Image or a Dictionary { "size", "color_mode", etc. }
        \~chinese 
        重新定义缓存数据
        @param newFrame: 图像 或 缓存字典 { "size", "color_mode", etc. }
        """
        raise NotImplementedError

    def changeBufferColorMode(self, newColorMode = SS_COLOR_MODE_MONO):
        """!
        \~english 
        Change buffer color mode
        @param newColorMode: new color mode. it can be chosen: { SS_COLOR_MODE_MONO | SS_COLOR_MODE_RGB }
        \~chinese
        改变缓存色彩模式
        @param newColorMode: 新色彩模式。 可选值： SS_COLOR_MODE_MONO 或 SS_COLOR_MODE_RGB

        \~ \n 
        @see SS_COLOR_MODE_MONO
        @see SS_COLOR_MODE_RGB
        """
        self._initBuffer(newColorMode, None)

    def getDisplaySize(self):
        return (0,0) if self._display_size==None else self._display_size

    def getBufferSize(self):
        raise NotImplementedError


# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
#    Base Display
#    by Kunpeng Zhang
#    v1.0.0    2018.3.20
#

import RPi.GPIO as GPIO

class RPiDiaplay:
    """!
    RPiDiaplay is a hardware abstraction of the display，
    You need to create a subclass from inherit it and 
    use the new subclass implement initialization and 
    operations of display chip.
    """
    width = None
    height = None

    # SPI interface
    _spi        = None
    _spi_mosi   = None
    _spi_dc     = None
    _spi_cs     = None
    _spi_reset  = None
    _spi_clk    = None

    # display buffer
    _buffer = None

    def _command(self, commands):
        """!
        Send command to hardware bus of display chip ( I2C, SPI, others )
        waitting for subclasses implement
        """
#         """Send command to spi bus of display chip, most DC pin need set to LOW """
#         if self._spi == None: raise "Do not setting SPI"
#         GPIO.output( self._spi_dc, 0 )
#         self._spi.writebytes( commands )
        raise NotImplementedError

    def _data(self, data):
        """!
        Send data to hardware bus of display chip ( I2C, SPI, others )
        waitting for subclasses implement
        """
#         """Send data to spi bus of display chip, most DC pin need set to HIGH """
#         if self._spi == None: raise "Do not setting SPI"
#         GPIO.output( self._spi_dc, 1 )
#         self._spi.writebytes( data )
        raise NotImplementedError

    def _init_config(self, width, height, spi=None, spiMosi= None, spiDC=None, spiCS=None, spiReset=None, spiClk=None):
        """!
        SPI hardware and display width, height initialization.
        """
        self._spi = spi
        self._spi_mosi = spiMosi
        self._spi_dc = spiDC
        self._spi_cs = spiCS
        self._spi_reset = spiReset
        self._spi_clk = spiClk

        self.width = width
        self.height = height
        
    def _init_io(self):
        """!
        GPIO initialization.
        Set GPIO into BCM mode and init other IOs mode
        """
        GPIO.setwarnings(False)
        GPIO.setmode( GPIO.BCM )
        pins = [ self._spi_dc ]
        for pin in pins:
            GPIO.setup( pin, GPIO.OUT )

    def _init_display(self):
        """!
        Display hardware initialization.
        waitting for subclasses implement
        """
        raise NotImplementedError

    def __init__ ( self, width, height, spi=None, spiMosi= None, spiDC=None, spiCS=None, spiReset=None, spiClk=None ):
        """!
        Initialize the RPiDiaplay object instance
        and config GPIO and others
        """
        self._init_config(width, height, spi, spiMosi, spiDC, spiCS, spiReset, spiClk)

    def clear(self, fill = 0x00):
        """!
        Clear buffer data and other data
        RPiDiaplay object just implemented clear buffer data
        """
        self._buffer = [ fill ] * ( self.width * self.height )

    def on(self):
        """!
        Power on display.
        waitting for subclasses implement
        """
        raise NotImplementedError

    def off(self):
        """!
        Power off display.
        waitting for subclasses implement
        """
        raise NotImplementedError

    def init(self):
        """!
        Change contrast of display.
        waitting for subclasses implement
        """
        raise NotImplementedError

    def reset(self):
        """!
        Reset display.
        waitting for subclasses implement
        """
        raise NotImplementedError
    
    def setContrast(self, contrast):
        """!
        Change contrast of display.
        waitting for subclasses implement
        """
        raise NotImplementedError
    
    def setBrightness(self, brightness):
        """!
        Change brightness of display.
        waitting for subclasses implement
        """
        raise NotImplementedError
    
    def display(self, buffer = None):
        """!
        Send an buffer data to display.
        waitting for subclasses implement
        """
        raise NotImplementedError
    
    def setImage(self, image):
        """!
        Set an image to display. the image can be PIL Image object or other image object.
        waitting for subclasses implement
        """
        raise NotImplementedError
        


# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
#    RPi-Spark Tone
#    @author Kunpeng Zhang
#    v1.0.0    2018.5.29
#

import RPi.GPIO as GPIO
from time import sleep
from time import time

TONE_A = 0
TONE_B = 1
TONE_C = 2
TONE_D = 3
TONE_E = 4
TONE_F = 5
TONE_G = 6

# /*
#     1b    2b    3b    4b    5b    6b    7b
# A    221    248    278    294    330    371    416
# B    248    278    294    330    371    416    467
# C    131    147    165    175    196    221    248
# D    147    165    175    196    221    248    278
# E    165    175    196    221    248    278    312
# F    175    196    221    234    262    294    330
# G    196    221    234    262    294    330    371
# */
TONE_BASS = [
  [221,248,278,294,330,371,416],
  [248,278,294,330,371,416,467],
  [131,147,165,175,196,221,248],
  [147,165,175,196,221,248,278],
  [165,175,196,221,248,278,312],
  [175,196,221,234,262,294,330],
  [196,221,234,262,294,330,371]
]

# /*
#     1      2     3     4     5     6     7
# A    441    495    556    589    661    742    833
# B    495    556    624    661    742    833    935
# C    262    294    330    350    393    441    495
# D    294    330    350    393    441    495    556
# E    330    350    393    441    495    556    624
# F    350    393    441    495    556    624    661
# G    393    441    495    556    624    661    742
# */
TONE_MID = [
  [441,495,556,589,661,742,833],
  [495,556,624,661,742,833,935],
  [262,294,330,350,393,441,495],
  [294,330,350,393,441,495,556],
  [330,350,393,441,495,556,624],
  [350,393,441,495,556,624,661],
  [393,441,495,556,624,661,742]
]

# /*
#     1#    2#    3#    4#    5#    6#    7#
# A  882  990    1112  1178  1322  1484  1665
# B  990  1112   1178  1322  1484  1665  1869
# C  525  589    661    700  786    882  990
# D  589  661    700    786  882    990  1112
# E  661  700    786    882  990    1112  1248
# F  700  786    882    935  1049  1178  1322
# G  786  882    990    1049  1178  1322  1484
# */
TONE_TREBLE = [
  [882,990,1112,1178,1322,1484,1665],
  [990,1112,1178,1322,1484,1665,1869],
  [525,589,661,700,786,882,990],
  [589,661,700,786,882,990,1112],
  [661,700,786,882,990,1112,1248],
  [700,786,882,935,1049,1178,1322],
  [786,882,990,1049,1178,1322,1484]
]

class RPiTonePlayer:
    """!
    \~english Use PWM to play tones
    \~chinese 使用PWM播放音符
    """
    TONE_DUTY = None
    MUTE_TONE = 0

    _pinPWM = None
    _pwmPlayer = None

    def __init__(self, pinPWM = None, toneDuty = 20):
        """!
        \~english
        Initialize the RPiTonePlayer object instance.
        @param pinPWM: This is a GPIO number of PWM for output
        @param toneDuty: The width of the pulse, 
                         its value can be chosen: 5 ~ 60
        @note on Raspberry Pi pwm output io you can be chosen in (BCM Mode): 12, 13, 18

        \~chinese
        初始化 RPiTonePlayer 对象实例
        @param pinPWM: 输出的 PWM 的 GPIO IO 编号
        @param toneDuty: 控制脉冲的宽度, 取值： 5 ~ 50
        @note 树莓派 pwm 输出可以选择（BCM模式）：12,13,18
        """
        self._pinPWM = pinPWM

        if toneDuty>50:
            self.TONE_DUTY = 50
        elif toneDuty<5:
            self.TONE_DUTY = 5
        else:
            self.TONE_DUTY = toneDuty

        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._pinPWM, GPIO.OUT)

    def _initPWMPlayer(self, freq):
        self._pwmPlayer = GPIO.PWM(self._pinPWM, freq)  # channel=12 frequency=50Hz
        #self._pwmPlayer.start(self.TONE_DUTY)
        #self._pwmPlayer.stop()

    def _delay(self, delay = 0):
        startTime=time()
        while True: 
            if (time() - startTime) > delay: break

    def stopTone(self):
        """!
        \~english Stop tone play
        \~chinese Stop tone play
        """
        if self._pwmPlayer == None: return False
        self._pwmPlayer.stop()

    def playTone(self, freq, reps = 1, delay = 0.1, muteDelay = 0.0):
        """!
        \~english Play a tone
        \~chinese 播放音符
        \~english 
        @param freq
        @param reps
        @param delay >= 0(s) if 0 means do not delay. tone play will be Stop immediately <br>
        @param muteDelay >= 0(s) If 0 means no pause after playing, play the next note immediately
        \~chinese 
        @param freq: 频率
        @param reps: 重复次数
        @param delay >= 0(s) 如果是 0 意味着不延迟。 音符会立即停止播放 <br>
        @param muteDelay >= 0(s) 如果是 0 表示音符播放结束后没有停顿，立刻播放下一个音符
        """
        if freq == 0: 
            self.stopTone()
            self._delay(delay)
            #sleep(delay)
            return False

        if self._pwmPlayer == None: self._initPWMPlayer(freq)

        for r in range(0,reps):
            self._pwmPlayer.start(self.TONE_DUTY)
            self._pwmPlayer.ChangeFrequency( freq )
            self._delay(delay)
            #sleep(delay)

            if muteDelay>0:
                self.stopTone()
                self._delay(muteDelay)
                #sleep(muteDelay)

        return True

    def playToneList(self, playList = None):
        """!
        \~english
        Play tone from a tone list
        @param playList a array of tones
        
        \~chinese
        播放音调列表
        @param playList: 音调数组
        
        \~english @note <b>playList</b> format:\n
        \~chinese @note <b>playList</b> 格式:\n
        \~
            <pre>
            [
              {"freq": 440, "reps": 1, "delay": 0.08, "muteDelay": 0.15},
              {"freq": 567, "reps": 3, "delay": 0.08, "muteDelay": 0.15},
              ...
            ]
           </pre>\n
        \~english
        \e delay: >= 0(s) if 0 means do not delay. tone play will be Stop immediately <br>
        \e muteDelay: 0.15 >= 0(s) If 0 means no pause after playing, play the next note immediately
        \~chinese
        \e delay: >= 0（s）如果是 0 意味着不延迟。 音调会立即停止播放 <br>
        \e muteDelay: >= 0（s）如果是 0 表示播放音符结束后没有停顿，立刻播放下一个音符
        """
        if playList == None: return False
        for t in playList:
            self.playTone(t["freq"], t["reps"], t["delay"], t["muteDelay"])

        self.stopTone()
        return True


# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
#    RPi OLED display ssd1306
#    @author Kunpeng Zhang
#    v1.0.0    2018.3.20
#

import RPi.GPIO as GPIO
import spidev
from PIL import Image
from .RPiDisplay import RPiDiaplay

class SSD1306Base( RPiDiaplay ):
    """!
    \~english SSD1306 Display Chip
    \~chinese SSD1306 显示芯片驱动
    """
    CMD_SSD1306_SET_MEM_ADDR_MODE       = 0x20
    CMD_SSD1306_SET_COLUMN_ADDR         = 0x21
    CMD_SSD1306_SET_PAGE_ADDR           = 0x22
    CMD_SSD1306_SET_DISPLAY_START_LINE  = 0x40  # 0x40(0) ~ 0x7F(63)
    CMD_SSD1306_SET_CONTRAST            = 0x81  # 0x00~0xFF
    CMD_SSD1306_SET_SEGMENT_REMAP_0     = 0xA0
    CMD_SSD1306_SET_SEGMENT_REMAP_1     = 0xA1
    CMD_SSD1306_ENTIRE_DISPLAY_ON_0     = 0xA4
    CMD_SSD1306_ENTIRE_DISPLAY_ON_1     = 0xA5
    CMD_SSD1306_NORMAL_DISPLAY          = 0xA6
    CMD_SSD1306_INVERSE_DISPLAY         = 0xA7
    CMD_SSD1306_SET_MULTIPLEX_RATIO     = 0xA8
    CMD_SSD1306_DISPLAY_OFF             = 0xAE
    CMD_SSD1306_DISPLAY_ON              = 0xAF
    CMD_SSD1306_CHARGE_PUMP             = 0x8D
    # CMD_SSD1306_SET_START_ADDR_PAGE_ADDR_MODE
    CMD_SSD1306_SCAN_DIRECTION_INC      = 0xC0
    CMD_SSD1306_SCAN_DIRECTION_DEC      = 0xC8
    CMD_SSD1306_SET_DISPLAY_OFFSET      = 0xD3
    CMD_SSD1306_SET_CLOCK_DIVIDE_RATIO  = 0xD5
    CMD_SSD1306_SET_PRECHARGE           = 0xD9
    CMD_SSD1306_SET_COM_PINS            = 0xDA
    CMD_SSD1306_SET_DESELECT_LEVEL      = 0xDB
    
    #Graphic Acceleration Command
    
    # This command stops the motion of scrolling. After sending 2Eh command to 
    # deactivate the scrolling action, the ram data needs to be rewritten. 
    CMD_SSD1306_SET_SCROLL_DEACTIVE     = 0x2E
    
    # This command starts the motion of scrolling and should only be issued after 
    # the scroll setup parameters have been defined by the scrolling setup 
    # commands:26h/27h/29h/2Ah . The setting in the last scrolling setup
    # command overwrites the setting in the previous scrolling setup commands.
    CMD_SSD1306_SET_SCROLL_ACTIVE       = 0x2F
    
    # This command consists of consecutive bytes to set up the horizontal scroll 
    # parameters and determines the scrolling start page, end page and scrolling
    # speed.
    CMD_SSD1306_SET_SCROLL_HORIZONTAL_RIGHT  = 0x26
    CMD_SSD1306_SET_SCROLL_HORIZONTAL_LEFT = 0x27
    
    # This command consists of 6 consecutive bytes to set up the continuous 
    # vertical scroll parameters and determines the scrolling start page, end 
    # page, scrolling speed and vertical scrolling offset. 
    # The bytes B[2:0], C[2:0] and D[2:0] of command 29h/2Ah are for the setting 
    # of the continuous horizontal scrolling. The byte E[5:0] is for the setting 
    # of the continuous vertical scrolling offset. All these bytes together are 
    # for the setting of continuous diagonal (horizontal + vertical) scrolling. 
    # If the vertical scrolling offset byte E[5:0] is set to zero, then only 
    # horizontal scrolling is performed (like command 26/27h).
    CMD_SSD1306_SET_SCROLL_HORIZONTAL_VERTICAL_RIGHT = 0x29
    CMD_SSD1306_SET_SCROLL_HORIZONTAL_VERTICAL_LEFT  = 0x2A
    
    # This command consists of 3 consecutive bytes to set up the vertical scroll 
    # area. For the continuous vertical
    # scroll function (command 29/2Ah), the number of rows that in vertical 
    # scrolling can be set smaller or equal to the MUX ratio.
    CMD_SSD1306_SET_SCROLL_VERTICAL_AREA= 0xA3
    
    ## 
    # \~english mirror horizontal (boolean)
    # \~chinese 水平镜像 (boolean)
    _mirror_h = None
    ## 
    # \~english mirror vertical (boolean)
    # \~chinese 垂直镜像 (boolean)
    _mirror_v = None
    ## 
    # \~english buffer page (int)
    # \~chinese 显存分页 (int)
    _mem_pages = None

    def _command(self, commands):
        """!
        \~english
        Send command to ssd1306, DC pin need set to LOW
        @param commands: an byte or array of bytes

        \~chinese 
        发送命令给 SSD1306，DC 需要设定为低电平 LOW
        @param commands: 一个字节或字节数组
        """
        if self._spi == None: raise "Do not setting SPI"
        GPIO.output( self._spi_dc, 0 )
        self._spi.writebytes( commands )

    def _data(self, data):
        """!
        \~english
        Send data to ssd1306, DC pin need set to HIGH
        @param data: sent to display chip of data. it can be an byte or array of bytes

        \~chinese
        发送数据给 SSD1306, DC 需要设定为高电平 HIGH
        @param data: 送到显示芯片的数据。 可以是一个字节或字节数组
        """
        if self._spi == None: raise "Do not setting SPI"
        GPIO.output( self._spi_dc, 1 )
        self._spi.writebytes( data )

    def _display_buffer(self, buffer ):
        """!
        \~english
        Send buffer data to physical display.
        @param buffer: sent to display chip of data.

        \~chinese
        将缓冲区数据发送到物理显示。
        @param buffer: 送到显示芯片的数据。
        """
        self._command([
            self.CMD_SSD1306_SET_COLUMN_ADDR,
            0, self.width-1,
            self.CMD_SSD1306_SET_PAGE_ADDR,
            0, self._mem_pages - 1
            ])
        self._data( buffer )

    def _init_display(self):
        """!
        \~english
        Initialize the SSD1306 display chip

        \~chinese
        初始化SSD1306显示芯片
        """
        self._command([
            # 0xAE
            self.CMD_SSD1306_DISPLAY_OFF,
            #Stop Scroll
            self.CMD_SSD1306_SET_SCROLL_DEACTIVE,
            # 0xA8 SET MULTIPLEX 0x3F
            self.CMD_SSD1306_SET_MULTIPLEX_RATIO,
            0x3F,
            # 0xD3 SET DISPLAY OFFSET
            self.CMD_SSD1306_SET_DISPLAY_OFFSET,
            0x00,
            # 0x40 Set Mapping RAM Display Start Line  (0x00~0x3F)
            self.CMD_SSD1306_SET_DISPLAY_START_LINE,
            # 0xDA Set COM Pins hardware configuration, (0x00/0x01/0x02)
            self.CMD_SSD1306_SET_COM_PINS,
            (0x02 | 0x10),
            self.CMD_SSD1306_SET_CONTRAST,
            0x7F,
            # 0xA4 Disable Entire Display On
            self.CMD_SSD1306_ENTIRE_DISPLAY_ON_0,
            # 0xA6 Set normal display
            self.CMD_SSD1306_NORMAL_DISPLAY,
            # 0xA7 Set inverse display
            # CMD_SSD1306_INVERSE_DISPLAY, 
            # 0xD5 Set osc frequency 0x80
            self.CMD_SSD1306_SET_CLOCK_DIVIDE_RATIO,
            0x80,
            # 0x8D Enable DC/DC charge pump regulator 0x14
            self.CMD_SSD1306_CHARGE_PUMP,
            0x14,
            # 0x20 Set Page Addressing Mode (0x00/0x01/0x02)
            self.CMD_SSD1306_SET_MEM_ADDR_MODE,
            0x01,
  
            # 0xC0 / 0xC8 Set COM Output Scan Direction
            #CMD_SSD1306_SCAN_DIRECTION_INC,
            #CMD_SSD1306_SCAN_DIRECTION_DEC,
            self.CMD_SSD1306_SCAN_DIRECTION_INC if self._mirror_v else self.CMD_SSD1306_SCAN_DIRECTION_DEC,
  
            # 0xA0 / oxA1 Set Segment re-map
            # 0xA0    left to right
            # 0xA1    right to left
            self.CMD_SSD1306_SET_SEGMENT_REMAP_0 if self._mirror_h else self.CMD_SSD1306_SET_SEGMENT_REMAP_1,
        ])

    def init(self):
        """!
        \~english
        Initialize the SSD1306 display chip, and ready for show something
        \~chinese
        初始化SSD1306显示芯片，并准备好接收显示命令和数据
        """
        self._init_io()
        self.reset()
        self._init_display()

    def clear(self, fill = 0x00):
        """!
        \~english
        Clear buffer data and fill color into buffer
        @param fill: a color value, it will fill into buffer.<br>
                The SSD1306 only chosen two colors: <br>
                   0 (0x0): black <br>
                   1 (0x1): white <br>

        \~chinese
        清除缓冲区数据并在缓冲区中填充颜色
        @param fill: 一个颜色值，它会填充到缓冲区中 <br>
                     SSD1306只能选择两种颜色： <br>
                        0（0x0）：黑色 <br>
                        1（0x1）：白色 <br>
        """
        self._buffer = [ fill ] * ( self.width * self._mem_pages )

    def on(self):
        """!
        \~english power on display.
        \~chinese 开启显示屏
        """
        self._command( [self.CMD_SSD1306_DISPLAY_ON] )

    def off(self):
        """!
        \~english power off display.
        \~chinese 关闭显示屏
        """
        self._command( [self.CMD_SSD1306_DISPLAY_OFF] )

    def reset(self):
        """!
        \~english Reset display
        \~chinese 复位显示屏
        """
        if self._spi_reset == None: return
        GPIO.output( self._spi_reset, 1 )
        time.sleep(0.002)
        GPIO.output( self._spi_reset, 0 )
        time.sleep(0.015)
        GPIO.output( self._spi_reset, 1 )

    def setContrast(self, contrast):
        """!
        \~english 
        Change contrast of SSD1306 display chip
        @param contrast: a contrast value, the range is: 0-255
        
        \~chinese 
        改变 SSD1306 显示屏对比度
        @param contrast: 对比度值，范围是：0-255
        """
        self._command([self.CMD_SSD1306_SET_CONTRAST, contrast])
        
    def setBrightness(self, brightness):
        """!
        \~english 
        @deprecated SSD1306 do not supported brightness justment
        \~chinese 
        @deprecated SSD1306不支持亮度调整
        """
        raise ValueError("Has not brightness controller")

    def display(self, buffer = None):
        """!
        \~english 
        Write buffer to physical display.
        @param buffer: Data to display，If <b>None</b> mean will use self._buffer data to display
        \~chinese 
        将缓冲区写入物理显示屏。
        @param buffer: 要显示的数据，如果是 <b>None</b>(默认) 将把 self._buffer 数据写入物理显示屏 
        """
        if buffer != None:
            self._display_buffer( buffer )
        else:
            self._display_buffer( self._buffer )
            
    def setImage(self, image):
        """!
        \~english
        Convert image to the buffer, The image mode must be 1 and image size 
        equal to the display size image type is Python Imaging Library image.
        @param image: a PIL image object

        \~chinese
        将图像转换为缓冲区，这个图像的色彩模式必须为 1 同时图像大小必须等于显存大小，
        图像类型： PIL Image (Python Imaging Library)
        @param image: PIL图像对象

        \n \~
        @note
        <pre>
        ssd1306.setImage( aPILImage )
        ssd1306.display()
        </pre>
        """
        if image.mode != '1':
            raise ValueError('The image color must be in mode \"1\".')

        imgWidth, imgHeight = image.size
        if imgWidth != self.width or imgHeight != self.height:
            raise ValueError('The image must be same dimensions as display ( {0} x {1} ).' \
                .format(self.width, self.height))

        # First then shift left 
        pixByte = [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80]
        bi = 0
        pixs = image.load()
        for x in range( 0, self.width ):
            for y in range( 0, self.height, 8 ):
                pixBits = 0x00
                # first then range(8)
                for py in [0,1,2,3,4,5,6,7]:
                    pixBits |= (0x00 if pixs[x, y+py] == 0 else pixByte[py])
                self._buffer[bi] = pixBits
                bi += 1

    def scrollOn(self):
        """!
        \~english 
        Start scroll screen
        use: SSD1306Base#scrollWith to set up scroll mode
        \~chinese 
        开始滚屏
        使用：SSD1306Base#scrollWith 设置滚动模式
        """
        self._command([self.CMD_SSD1306_SET_SCROLL_ACTIVE])
        
    def scrollOff(self):
        """!
        \~english Stop scroll screen
        \~chinese 停止滚屏
        
        \n \~
        @see scrollOn
        """
        self._command([self.CMD_SSD1306_SET_SCROLL_DEACTIVE])

    def scrollWith(self, hStart = 0x00, hEnd=0x00, vOffset = 0x00, vStart=0x00, vEnd=0x00, int = 0x00, dire = "left" ):
        """!
        \~english 
        Scroll screen
        @param hStart: Set horizontal scroll PAGE start address, value can be chosen between 0 and 7
        @param hEnd:  Set horizontal scroll PAGE end address, value can be chose between 0 and 7
        @param vOffset: Vertical scroll offset row, if set to 0x00(0) means off vertical scroll
        @param vStart: Vertical scroll start line, value can be chose between 0x00 and screen.height
        @param vEnd: Vertical scroll end line, value can be chose between 0x00 and screen.height
        @param int: Set time interval between each scroll step
        @param dire: Set scroll direction, value can be "left" or "right"
        
        \~chinese
        屏幕滚动
        @param hStart: 设置水平滚动PAGE起始地址，数值可以在0到7之间选择
        @param hEnd:  设置水平滚动PAGE结束地址，值可以在0和7之间选择
        @param vOffset: 垂直滚动偏移行，如果设置为0x00（0），表示关闭垂直滚动
        @param vStart: 垂直滚动起始行，值可以在0x00和screen.height之间选择
        @param vEnd: 垂直滚动结束行，值可以在0x00和screen.height之间选择
        @param int: 设置每个滚动步骤之间的时间间隔
        @param dire: 设置滚动方向，数值可以是 "left" 或 "right"
        """
        self._command( [self.CMD_SSD1306_SET_SCROLL_DEACTIVE] )
        if vOffset != 0:
            self._command( [
                    self.CMD_SSD1306_SET_SCROLL_VERTICAL_AREA,
                    vStart,
                    vEnd,
                    0x00
                ])

        self._command( [
            self.CMD_SSD1306_SET_SCROLL_HORIZONTAL_VERTICAL_LEFT if dire.upper()=="LEFT" else self.CMD_SSD1306_SET_SCROLL_HORIZONTAL_VERTICAL_RIGHT,
            0x00, 
            hStart, 
            int, 
            hEnd, 
            vOffset, 
            0x00,
            self.CMD_SSD1306_SET_SCROLL_ACTIVE
            ])


class SSD1306_128x64(SSD1306Base):
    """!
    \~english 
    128x64 OLED of SSD1306 Display Chip
    uses: SPI

    \~chinese
    SSD1306 128x64 OLED 显示芯片驱动, 使用 SPI 总线
    """
    def __init__( self, spi=None, spiMosi= None, spiDC=None, spiCS=None, spiReset=None, spiClk=None, mirrorH = 0, mirrorV = 0  ):
        """!
        \~english Initialize the SSD1306 128x64 OLED display object instance and config GPIO and others
        \~chinese 实例化 SSD1306 128x64 OLED 显示屏对象，同时初始化 GPIO 和其他

        \~english
        @param spi: a spi bus object instance
        @param spiMosi: spi mosi io number, eg. 10 default is None
        @param spiDC: spi dc io number, eg. 9 default is None
        @param spiCS: spi cs io number, eg. 8 default is None
        @param spiReset: spi reset io number, eg. 18 default is None. <br>
                        If <b>None</b> mean this chip do not need reset
        @param spiClk: spi clk io number, eg. 11 default is None
        @param mirrorH: The displayed image flips horizontal
        @param mirrorV: The displayed image flips vertically
        
        \~chinese
        @param spi: SPI 总线对象实例
        @param spiMosi: SPI MOSI IO, 例如： 10, 默认: None
        @param spiDC: SPI DC IO, 例如： 9, 默认: None
        @param spiCS: SPI CS IO, 例如： 8, 默认: None
        @param spiReset: SPI RESET IO, 例如： 18, 默认: None <br>
                        如果是 <b>None</b> 停用 SPI RESET 功能
        @param spiClk: SPI CLK IO, 例如：11， 默认 None
        @param mirrorH: 显示屏水平镜向显示
        @param mirrorV: 显示屏垂直镜向显示

        \~
        @note
        \~english A siample for use SSD1306_128x64 below:
        \~chinese 使用 SSD1306_128x64 示例：
        \~\n
        <hr>
        <pre>
        import spidev
        from PIL import Image
        from JMRPiSpark.Drives.Display.SSD1306 import SSD1306_128x64 \n
        class CONFIG_DSP:
            DSP_RESET       = None
            DSP_DC          = 9
            DSP_SPI_PORT    = 0
            DSP_SPI_DEVICE  = 0
            DSP_SPI_MAX_SPEED_HZ = 2000000
            DSP_MIRROR_H    = True
            DSP_MIRROR_V    = True \n
        spi = spidev.SpiDev()
        spi.open( CONFIG_DSP.DSP_SPI_PORT, CONFIG_DSP.DSP_SPI_DEVICE)
        spi.max_speed_hz = CONFIG_DSP.DSP_SPI_MAX_SPEED_HZ
        spi.cshigh = False
        spi.mode = 0 \n
        myDSP = SSD1306_128x64 ( 
            spi,  
            spiDC = CONFIG_DSP.DSP_DC,
            spiReset = CONFIG_DSP.DSP_RESET,
            mirrorH = CONFIG_DSP.DSP_MIRROR_H, 
            mirrorV = CONFIG_DSP.DSP_MIRROR_V
            ) \n
        myDSP.init()
        myDSP.on()
        </pre>
        """
        self._init_config(128, 64, spi, spiMosi, spiDC, spiCS, spiReset, spiClk)
        self._mirror_h = mirrorH
        self._mirror_v = mirrorV
        self._mem_pages = 64 //8
        self._buffer = [0] * ( 128 * self._mem_pages )

# class SSD1306_128x32(SSD1306Base):
#     def __init__( self, spi=None, spiMosi= None, spiDC=None, spiCS=None, spiReset=None, spiClk=None, mirrorH = 0, mirrorV = 0  ):        
#         self._init_config(128, 32, spi, spiMosi, spiDC, spiCS, spiReset, spiClk)
#         self._mirror_h = mirrorH
#         self._mirror_v = mirrorV
#         self._mem_pages = 32 //8
#         self._buffer = [0] * ( 128 * self._mem_pages )
# 
# class SSD1306_64x32(SSD1306Base):
#     def __init__( self, spi=None, spiMosi= None, spiDC=None, spiCS=None, spiReset=None, spiClk=None, mirrorH = 0, mirrorV = 0  ):        
#         self._init_config(64, 32, spi, spiMosi, spiDC, spiCS, spiReset, spiClk)
#         self._mirror_h = mirrorH
#         self._mirror_v = mirrorV
#         self._mem_pages = 32 //8
#         self._buffer = [0] * ( 64 * self._mem_pages )



# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Kunpeng Zhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# #########################################################
#
#    RPi-Spark SSD1306Screen
#    by Kunpeng Zhang
#    v1.0.0
#
#    Using SSD1306 chip-driven OLED display
#    使用 SSD1306 芯片驱动的 OLED 显示屏

from .SSPILScreen import SSPILScreen

class SScreenSSD1306( SSPILScreen ):
    """!
    \~english
    Using SSD1306 chip to implement the SScreen object
    Supported: JMRPiDisplay SSD1306 driver and Adafruit SSD1306 driver

    \~chinese
    使用 SSD1306 显示芯片实现的 SScreen 对象
    支持: JMRPiDisplay_SSD1306 和 Adafruit SSD1306 driver
    """
    def __init__(self, display, bufferColorMode, bufferSize=None, displayDirection=0 ):
        """!
        \~english
        Initialize the SScreenSSD1306 object instance
        @param display: a display hardware instance, eg. SSD1306
        @param bufferColorMode: color mode, can be chosen: SS_COLOR_MODE_MONO ("1") or SS_COLOR_MODE_RGB ("RGB")
        @param bufferSize: size of buffer, eg. (128,64) or (320, 240) ...
        @param displayDirection: direction of display, can be: 0, 90, 180, 270
        
        \~chinese
        初始化 SScreenSSD1306 对象实例
        @param display: 显示屏硬件对象实例，例如： SSD1306
        @param bufferColorMode: 缓存色彩模式，取值： SS_COLOR_MODE_MONO ("1") 或 SS_COLOR_MODE_RGB ("RGB")
        @param bufferSize: 缓存大小，例如： (128,64) or (320, 240) ...
        @param displayDirection: 显示屏方向，取值：0, 90, 180, 270
        """
        self._checkBufferColorMode(bufferColorMode)
        # Initialize display.
#         display.begin()
        self._display_color_mode = "1"
        self._initDisplay(display, displayDirection, (display.width, display.height))

        # Initialize buffer and canvas
        self._initBuffer( bufferColorMode, bufferSize )

    def refresh(self):
        """!
        \~english
        Update current view content to display
        Supported: JMRPiDisplay_SSD1306 and Adafruit SSD1306 driver

        \~chinese
        更新当前视图内容到显示屏
        支持: JMRPiDisplay_SSD1306 和 Adafruit SSD1306 driver
        """
        try:
            # suport for RPiDisplay SSD1306 driver
            self.Display.setImage( self._catchCurrentViewContent() )
        except:
            try:
                # suport for Adafruit SSD1306 driver
                self.Display.image( self._catchCurrentViewContent() )
            except:
                raise "Can not update image to buffer."

        self.Display.display()


#
# RPi-Spark pHAT Drives
# Author: Kunpeng Zhang
# 2018.6.6
#

from setuptools import setup, find_packages

classifiers = [
	'Development Status :: 4 - Beta',
	'Operating System :: POSIX :: Linux',
	'License :: OSI Approved :: MIT License',
	'Intended Audience :: Developers',
	'Programming Language :: Python :: 2.7',
	'Programming Language :: Python :: 3',
	'Topic :: Software Development',
	'Topic :: System :: Hardware',
	'Topic :: System :: Hardware :: Hardware Drivers'
]

keywords = (
	"development kit"
	"oled"
	"monochrome greyscale color"
	"ssd1306"
	"mpu6050"
	"attitude shake motion gyroscope accelerometer thermometer sensor"
	"spi i2c 128x64"
	"key buttons"
	"joystick game"
	"audio speaker headset headphone earphone"
	"pwm tone"
	"ogg mp3 wave"
	"gpio extended pads"
)

desc = 'The Raspberry Pi Spark pHat board let you to easily develop interesting applications use the GPIO of Raspberry Pi. It included SSD1306 128x64 OLED, MPU6050 Sensor (Gyroscope, Accelerometer, Thermometer Sensor), 5 ways joystick, 2 action keys, 3.5mm stereo headphone jack, Speaker and 20 extended GPIO pads'

setup (
	name              = 'JMRPi.Spark',
	version           = '1.0.6',
	author            = 'Kunpeng Zhang',
	author_email      = 'support@mobinrg.com',
	description       = desc,
	long_description  = desc,
	platforms		  = ['Linux'],
	license           = 'MIT',
	classifiers        = classifiers,
	keywords	  	  = keywords,
	url               = 'https://github.com/mobinrg/rpi_spark_drives',
	dependency_links  = [],
	install_requires  = [],
	packages          = find_packages()
)




from .SSPILScreen import SSPILScreen

class SScreenILI9341( SSPILScreen ):
    """This class work with PIL Lib.
    """

    def __init__(self, display, bufferColorMode, bufferSize=None, displayDirection=0 ):
        self._checkBufferColorMode(bufferColorMode)
        # Initialize display.
#         display.begin()
        self._display_color_mode = "RGB"
        self._initDisplay(display, displayDirection, (display.width, display.height))

        # Initialize buffer and canvas
        self._initBuffer( bufferColorMode, bufferSize )
        pass

    def refresh(self):
        """Update current view content to display
        """
        self.Display.display( self._catchCurrentViewContent() )
        pass

    def clear(self):
        """Clear display and screen's canvas
        """
        self.clearCanvas()
        self.Display.clear()
        pass

    pass